@@ -141,13 +141,14 @@ class closing(object):
         self.thing.close()
 
 
+# Inspired by discussions on http://bugs.python.org/issue13585
 class ContextStack(object):
-    """Context for programmatic management of resource cleanup
+    """Context manager for dynamic management of a stack of exit callbacks
     
     For example:
     
         with ContextStack() as stack:
-            files = [stack.enter_context(fname) for fname in filenames]
+            files = [stack.enter_context(open(fname)) for fname in filenames]
             # All opened files will automatically be closed at the end of
             # the with statement, even if attempts to open files later
             # in the list throw an exception
@@ -155,13 +156,36 @@ class ContextStack(object):
     """
     def __init__(self):
         self._callbacks = deque()
+        
+    def preserve(self):
+        """Preserve the context stack by transferring it to a new instance"""
+        new_stack = type(self)()
+        new_stack._callbacks = self._callbacks
+        self._callbacks = deque()
+        return new_stack
 
+    def _register_cm_exit(self, cm, cm_exit):
+        """Helper to correctly register callbacks to __exit__ methods"""
+        def _exit_wrapper(*exc_details):
+            return cm_exit(cm, *exc_details)
+        _exit_wrapper.__self__ = cm
+        self.register_exit(_exit_wrapper)
+        
     def register_exit(self, callback):
         """Registers a callback with the standard __exit__ method signature
 
         Can suppress exceptions the same way __exit__ methods can.
+
+        Also accepts any object with an __exit__ method (registering the
+        method instead of the object itself)
         """
-        self._callbacks.append(callback)
+        _cb_type = type(callback)
+        try:
+            exit = _cb_type.__exit__
+        except AttributeError:
+            self._callbacks.append(callback)
+        else:
+            self._register_cm_exit(callback, exit)
         return callback # Allow use as a decorator
 
     def register(self, callback, *args, **kwds):
@@ -169,9 +193,12 @@ class ContextStack(object):
         
         Cannot suppress exceptions.
         """
-        def _wrapper(exc_type, exc, tb):
+        def _exit_wrapper(exc_type, exc, tb):
             callback(*args, **kwds)
-        return self.register_exit(_wrapper)
+        # We changed the signature, so using @wraps is not appropriate, but
+        # setting __wrapped__ may still help with introspection
+        _exit_wrapper.__wrapped__ = callback
+        self.register_exit(_exit_wrapper)
 
     def enter_context(self, cm):
         """Enters the supplied context manager
@@ -183,9 +210,7 @@ class ContextStack(object):
         _cm_type = type(cm)
         _exit = _cm_type.__exit__
         result = _cm_type.__enter__(cm)
-        def _exit_wrapper(*exc_details):
-            return _exit(cm, *exc_details)
-        self.register_exit(_exit_wrapper)
+        self._register_cm_exit(cm, _exit)
         return result
 
     def close(self):
